"""Credential management

.. currentmodule:: datalad_next.credman
.. autosummary::
   :toctree: generated

   CredentialManager
   verify_property_names
"""
from .manager import (
    CredentialManager,
    verify_property_names,
)
