"""Configuration query and manipulation

This modules provides the central ``ConfigManager`` class.

.. currentmodule:: datalad_next.config
.. autosummary::
   :toctree: generated

   ConfigManager
"""

from datalad.config import ConfigManager
